
/****** coor.c **********************************************************/
/*									*/
/*   Included in this module are two recursive coordinate bisection 	*/
/*   program based on the median-of-medians algorithm. The programs	*/
/*   return the mapping array of parititoning.				*/
/*									*/
/*   RecCoorBisecMap:      Recursive coordinate bisection program.	*/
/*   WeighRecCoorBisecMap: Weighted recursive coordinate bisection.	*/
/*   									*/
/*************************************************** Shin 11/22/93 ******/

#include <stdio.h>
#include "parti.h"
#include "bisec.h"

#define  MAXPART        512     /* maximum number of partitions */
#define	 REMAPLEVEL	2	/* default value */

extern	double	*doublelist;	/* pointer to double array */

/* Recursive coordinate bisection */
void
RecCoorBisecMap (level, map, ndata, ndim, data)
int    level;			/* level of binary disection */
int    *map;			/* result mapping of the partition */
int    ndata;			/* number of nodes */
int    ndim;			/* number of dimensions */
double *data[];    		/* coordinate arrays */

{
    int    totalpart;		/* total number of partitions */
    int    npart;		/* current number of partitions */
    int    direction;		/* partitioning direction */
    int    ptr[MAXPART+1];	/* pointer of begin of partition */
    int    *proc;		/* processor assignments */
    int    *index;		/* index of each element */
    int    i, j, k, n, no;
    Val    *midlist;		/* array of medians */
    Val    *temp;

    /* initialize parameters */
    totalpart = (1 << level);

    /* allocate memory for necessary arrays*/
    index   = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
    proc    = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
    midlist = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
    temp    = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));

    /* put the coordinates into data array */
    for (i = 0; i <= ndata; i++) {
        index[i] = i;
	proc[i] = 0;
    }

    /* sort the arrays of each partition and find the median */
    for (i = 0, ptr[0] = 0; i < level; i++) {
        direction = i % ndim;		/* direction of partition */
        npart = (1 << i);		/* number of partitions */
        ptr[npart] = ndata;		/* number of nodes */

        /* sort the arrays of each partition */
	doublelist = data[direction];
        for (j = 0; j < npart; j++) {
            /* sort the array via coordinates */
            qsort (&index[ptr[j]], ptr[j+1]-ptr[j], sizeof(int), comparedouble);
        }

	/* apply the method median of medians to find
	   the median of the distributed array */
	find_median (level, npart, ptr, doublelist, index, proc, midlist, temp);
    }

    /* put the result of partition to map_array */
    for (i = 0; i < ndata; i++)
        map[index[i]] = proc[index[i]];

    /* release memory */
    free (index);
    free (proc);
    free (temp);
    free (midlist);
}

/* when the number of processor is large, the calls to CoorBisecMap()
   will be changed to CoorBisec() to reduce execution time. */
void
RecCoorBisecRemap (level, maparray, ndata, ndim, data)
int	level;			/* level of binary disection */
int     *maparray;              /* result of partitioning */
int     ndata;                  /* number of nodes */
int	ndim;			/* number of dimenstions */
double  *data[];		/* coordinates */

{
    int     start;          /* start index of local data */
    int     newndata;       /* number of data after partitioning */
    int     nlist[MAXPART]; /* list of ndata */
    int     temp[MAXPART];  /* temporary array */
    int     *myindex;       /* indexes of local elements */
    int     totalsize, localsize;
    int     *proc, *local;
    int     i, j;
    double  **newdata;
    TTABLE  *table;
    SCHED   *sched;

    /* find the global size */
    for (i = 0; i < PARTI_numnodes; i++)
        nlist[i] = 0;
    nlist[PARTI_mynode] = ndata;
    PARTI_gisum (nlist, PARTI_numnodes, temp);
    for (localsize = nlist[0], i = 1; i < PARTI_numnodes; i++) {
	if (nlist[i] > localsize)
	    localsize = nlist[i];
        nlist[i] += nlist[i-1];
    }
    totalsize = nlist[PARTI_numnodes-1];
    localsize = ((localsize << 1) > totalsize)? totalsize: (localsize << 1);

    /* allocate memory */
    myindex = (int *) malloc (localsize * sizeof(int));
    newdata = (double **) malloc (sizeof(double *) * ndim);
    for (i = 0; i < ndim; i++)
        newdata[i] = (double *) malloc (localsize * sizeof(double));

    /* generate indexes */
    start = (PARTI_mynode) ? nlist[PARTI_mynode-1]: 0;
    for (i = 0; i < ndata; i++)
        myindex[i] = start + i;

    /* build translation table */
    table = build_translation_table (BLOCKED, myindex, ndata);

    /* copy data arrays to new arrays */
    for (i = 0; i < ndim; i++)
        for (j = 0; j < ndata; j++) {
    	    newdata[i][j] = data[i][j];
    }

    /* call the binary disection */
    newndata = ndata;
    RecCoorBisec (level, REMAPLEVEL, myindex, &newndata, ndim, newdata);

    /* free allocated memory */
    for (i = 0; i < ndim; i++)
        free (newdata[i]);
    free (newdata);

    /* generate a schedule */
    proc  = (int *) malloc (newndata * sizeof(int));
    local = (int *) malloc (newndata * sizeof(int));
    dereference (table, myindex, local, proc, newndata);
    sched = schedule (local, proc, newndata, 1);

    /* produce maparray */
    for (i = 0; i < newndata; i++)
        proc[i] = PARTI_mynode;
    PARTI_scatter (sched, proc, maparray, sizeof(int), NULL);

    /* free allocated memory */
    free (local);
    free (proc);
    free (myindex);
    free_table (table);
    free_sched (sched);
}

/* weighted recursive coordiate bisection */
WeighRecCoorBisecMap (level, map, load, ndata, ndim, data)
int    level;			/* level of binary disection */
int    *map;			/* result mapping of the partition */
int    *load;			/* load of every vertex */
int    ndata;			/* number of nodes */
int    ndim;			/* number of dimensions */
double *data[];    		/* coordinate arrays */

{
    int    totalpart;		/* total number of partitions */
    int    npart;		/* current number of partitions */
    int    direction;		/* partitioning direction */
    int    ptr[MAXPART+1];	/* pointer of begin of partition */
    int    *index;		/* index of arrays */
    int    *proc;		/* processor assignments */
    int    *weight;		/* weight */
    int    i, j, k, n, no;
    Val    *midlist;		/* array of medians */
    Val    *temp;

    /* initialize parameters */
    totalpart = (1 << level);

    /* allocate memory for necessary arrays*/
    index   = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
    proc    = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
    midlist = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
    temp    = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
    /* statements for weighted partitioner */
    weight  = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));

    /* put the coordinates into data array */
    for (i = 0; i <= ndata; i++) {
	index[i] = i;
	proc[i] = 0;
    }

    /* sort the arrays of each partition and find the median */
    for (i = 0, ptr[0] = 0; i < level; i++) {
        direction = i % ndim;    /* direction of partition */
        npart = (1 << i);    /* number of partitions */
        ptr[npart] = ndata;    /* number of nodes */

        /* sort the arrays of each partition */
	doublelist = data[direction];
        for (j = 0; j < npart; j++) {
            /* sort the array via coordinates */
            qsort (&index[ptr[j]], ptr[j+1]-ptr[j], sizeof(int), comparedouble);
        }

	/* install the weight array */
	for (j = 1, weight[0] = 0; j <= ndata; j++)
	    weight[j] = weight[j-1] + load[index[j-1]];
	
	/* apply the method median of medians to find
	   the median of the distributed array */
	find_median_weight (level, npart, ptr, doublelist, weight, index, proc, midlist, temp);
    }

    /* put the result of partition to map_array */
    for (i = 0; i < ndata; i++) {
        map[index[i]] = proc[index[i]];
    }

    /* release memory */
    free (index);
    free (proc);
    free (temp);
    free (weight);
    free (midlist);
}

/* when the number of processor is large, the calls to WeighCoorBisecMap()
   will be changed to WeighCoorBisec() to reduce execution time. */
void
WeighRecCoorBisecRemap (level, maparray, load, ndata, ndim, data)
int	level;			/* level of binary disection */
int     *maparray;              /* result of partitioning */
int	*load;			/* loads of elements */
int     ndata;                  /* number of nodes */
int	ndim;			/* number of dimenstions */
double  *data[];		/* coordinates */

{
    int     start;          /* start index of local data */
    int     newndata;       /* number of data after partitioning */
    int     nlist[MAXPART]; /* list of ndata */
    int     temp[MAXPART];  /* temporary array */
    int     *myindex;       /* indexes of local elements */
    int     totalsize, localsize;
    int	    *newload;
    int     *proc, *local;
    int     i, j;
    double  **newdata;
    TTABLE  *table;
    SCHED   *sched;

    /* find the global size */
    for (i = 0; i < PARTI_numnodes; i++)
        nlist[i] = 0;
    nlist[PARTI_mynode] = ndata;
    PARTI_gisum (nlist, PARTI_numnodes, temp);
    for (localsize = nlist[0], i = 1; i < PARTI_numnodes; i++) {
	if (nlist[i] > localsize)
	    localsize = nlist[i];
        nlist[i] += nlist[i-1];
    }
    totalsize = nlist[PARTI_numnodes-1];
    localsize = ((localsize << 1) > totalsize)? totalsize: (localsize << 1);

    /* allocate memory */
    myindex = (int *) malloc (localsize * sizeof(int));
    newload = (int *) malloc (localsize * sizeof(int));
    newdata = (double **) malloc (sizeof(double *) * ndim);
    for (i = 0; i < ndim; i++)
        newdata[i] = (double *) malloc (localsize * sizeof(double));

    /* generate indexes */
    start = (PARTI_mynode) ? nlist[PARTI_mynode-1]: 0;
    for (i = 0; i < ndata; i++) {
        myindex[i] = start + i;
	newload[i] = load[i];
    }

    /* build translation table */
    table = build_translation_table (BLOCKED, myindex, ndata);

    /* copy data arrays to new arrays */
    for (i = 0; i < ndim; i++)
        for (j = 0; j < ndata; j++) {
    	    newdata[i][j] = data[i][j];
    }

    /* call the binary disection */
    newndata = ndata;
    WeighRecCoorBisec (level, REMAPLEVEL, myindex, newload, &newndata, ndim, newdata);

    /* free allocated memory */
    for (i = 0; i < ndim; i++)
        free (newdata[i]);
    free (newdata);

    /* generate a schedule */
    proc  = (int *) malloc (newndata * sizeof(int));
    local = (int *) malloc (newndata * sizeof(int));
    dereference (table, myindex, local, proc, newndata);
    sched = schedule (local, proc, newndata, 1);

    /* produce maparray */
    for (i = 0; i < newndata; i++)
        proc[i] = PARTI_mynode;
    PARTI_scatter (sched, proc, maparray, sizeof(int), NULL);

    /* free allocated memory */
    free (local);
    free (proc);
    free (newload);
    free (myindex);
    free_table (table);
    free_sched (sched);
}
